
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "branch": "master",
 "date": "2020-04-17T18:53:27+0100",
 "dirty": false,
 "error": null,
 "full-revisionid": "7f6feed543105843e5967c7ccbeee9badfb78f5a",
 "repository": "https://github.com/metawards/MetaWards",
 "version": "0.7.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
