"""
.. currentmodule:: metawards.extractors

Functions
=========

.. autosummary::
    :toctree: generated/

    extract_default
    output_default
    output_default_omp
    setup_output_default

"""

from ._extract_default import *
from ._output_default import *
